import base64
import getpass
import json
import os
from enum import Enum
from typing import Literal

import requests as r
from dapla import AuthClient
from dapla.auth import AuthError

from statbank.statbank_logger import logger


class UseDb(Enum):
    """Hold options for database choices, targetted at statbanken."""

    PROD = "PROD"
    TEST = "TEST"


class StatbankAuth:
    """Parent class for shared behavior between Statbankens "Transfer-API" and "Uttaksbeskrivelse-API"."""

    def __init__(self, use_db: UseDb | Literal["TEST", "PROD"] | None) -> None:
        """This init will never be used directly, as this class is always inherited from.

        This is for typing with Mypy.
        """
        if use_db is None:
            self.use_db = UseDb[self.check_env()]
        elif isinstance(use_db, UseDb):
            self.use_db = use_db
        else:
            self.use_db = UseDb[use_db]

    def _build_headers(self) -> dict[str, str]:
        return {
            "Authorization": self._build_auth(),
            "Content-Type": "multipart/form-data; boundary=12345",
            "Connection": "keep-alive",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept": r"*/*",
            "User-Agent": self._build_user_agent(),
        }

    @staticmethod
    def check_env() -> str:
        """Check if you are on Dapla or in prodsone.

        Simplified terribly by the addition of env vars for this, keeping this method for legacy reasons.

        Returns:
            str: "DAPLA" if on dapla, "PROD" if you are in prodsone.
        """
        return os.environ.get("DAPLA_ENVIRONMENT", "TEST")

    def check_database(self) -> str:
        """Checks if we are in prod environment. And which statbank-database we are sending to."""
        db = "TEST"
        if os.environ.get("DAPLA_ENVIRONMENT", "TEST") == "PROD":
            db = "PROD"
        if self.use_db:
            db = self.use_db.value
        return db

    def _build_user_agent(self) -> str:
        envir = os.environ.get("DAPLA_ENVIRONMENT", "TEST")
        service = os.environ.get("DAPLA_SERVICE", "JUPYTERLAB")
        region = os.environ.get("DAPLA_REGION", "ON_PREM")

        user_agent = f"{envir}-{region}-{service}-"

        return user_agent + r.utils.default_headers()["User-agent"]

    def _build_auth(self) -> str:
        username_encryptedpassword = (
            bytes(
                self._get_user(),
                "UTF-8",
            )
            + bytes(":", "UTF-8")
            + bytes(json.loads(self._encrypt_request().text)["message"], "UTF-8")
        )
        return "Basic " + base64.b64encode(username_encryptedpassword).decode("utf8")

    @staticmethod
    def _get_user() -> str:
        return getpass.getpass("Lastebruker:")

    def _use_test_url(self, test_env: str, prod_env: str) -> str:
        use_test = (
            self.use_db == UseDb.TEST
            and os.environ.get("DAPLA_ENVIRONMENT", "TEST") == "PROD"
        )
        env = test_env if use_test else prod_env
        return os.environ.get(env, f"Cant find {env} in environ.")

    def _encrypt_request(self) -> r.Response:
        db = self.check_database()
        try:
            headers = {
                "Content-type": "application/json",
            }
            if os.environ.get("DAPLA_REGION", "TEST") != "ON_PREM":
                headers["Authorization"] = f"Bearer {AuthClient.fetch_personal_token()}"
        except AuthError as err:
            logger.warning(str(err))
            headers = {
                "Content-type": "application/json",
            }

        return r.post(
            self._use_test_url("STATBANK_TEST_ENCRYPT_URL", "STATBANK_ENCRYPT_URL"),
            headers=headers,
            json={"message": getpass.getpass(f"Lastepassord ({db}):")},
            timeout=5,
        )

    def _build_urls(self) -> dict[str, str]:
        base_url = self._use_test_url("STATBANK_TEST_BASE_URL", "STATBANK_BASE_URL")
        end_urls = {
            "loader": "statbank/sos/v1/DataLoader?",
            "uttak": "statbank/sos/v1/uttaksbeskrivelse?",
            "gui": "lastelogg/gui/",
            "api": "lastelogg/api/",
        }
        return {k: base_url + v for k, v in end_urls.items()}



from __future__ import annotations

import datetime
import gc
import json
import os
import re
import urllib
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Literal
from typing import cast

import pandas as pd
import requests as r

from statbank.auth import StatbankAuth
from statbank.auth import UseDb
from statbank.globals import APPROVE_DEFAULT_JIT
from statbank.globals import OSLO_TIMEZONE
from statbank.globals import SSB_TBF_LEN
from statbank.globals import TOMORROW
from statbank.globals import Approve
from statbank.globals import _approve_type_check
from statbank.statbank_logger import logger

if TYPE_CHECKING:
    from statbank.api_types import TransferResultType


class StatbankTransfer(StatbankAuth):
    """Class for talking with the "transfer-API", which actually recieves the data from the user and sends it to Statbank.

    Attributes:
        data (dict[str, pd.DataFrame]):  name of "deltabell.dat" as keys.
            Number of DataFrames needs to match the number of "deltabeller" in
            the uttakksbeskrivelse.
            Dict-shape can be retrieved and validated before transfer with the
            Uttakksbeskrivelses-class.
        tableid (str): The numeric id of the table, matching the one found on the website.
            Should be a 5-length numeric-string. Alternatively it should be possible to send in the "hovedtabellnavn" instead of the tableid.
        shortuser (str): The abbrivation of username at ssb. Three letters, like "cfc"
        date (str): Date for publishing the transfer. Shape should be "yyyy-mm-dd", like "2022-01-01".
            Statbanken only allows publishing four months into the future?
        cc (str): First person to be notified by email of transfer. Defaults to the same as "shortuser"
        bcc (str): Second person to be notified by email of transfer. Defaults to the same as "cc"
        overwrite (bool):
            - False = no overwrite
            - True = overwrite
        approve (Approve | str | int):
            - 0 = MANUAL approval
            - 1 = AUTOMATIC approval at transfer-time (immediately)
            - 2 = JIT (Just In Time), approval right before publishing time
        use_db (UseDb | str | None):
            If you are in PROD-dapla and want to send to statbank test-database, set this to "TEST".
            When sending from TEST-environments you can only send to TEST-db, so this parameter is then ignored.
            Be aware that metadata tends to be outdated in the test-database.
        validation (bool):
            - True, if you want the python-validation code to run user-side.
            - False, if its slow and unnecessary.
        boundary (str): String that defines the splitting of the body in the transfer-post-request.
            Kept here for uniform choice through the class.
        urls (dict[str, str]): Urls for transfer, observing the result etc.,
            built from environment variables.
        headers (dict[str, str]): Might be deleted without warning.
            Temporarily holds the Authentication for the request.
        params (dict[str, str]): This dict will be built into the post request.
            Keep it in this nice shape for later introspection.
        body (str): The data parsed into the body-shape the Statbank-API expects in the transfer-post-request.
        response (requests.Response): The resulting response from the transfer-request. Headers might be deleted without warning.
    """

    def __init__(  # noqa: PLR0913
        self,
        data: dict[str, pd.DataFrame],
        tableid: str = "",
        shortuser: str = "",
        date: str | datetime.date | datetime.datetime = TOMORROW,
        cc: str = "",
        bcc: str = "",
        overwrite: bool = True,
        approve: int | str | Approve = APPROVE_DEFAULT_JIT,
        use_db: UseDb | Literal["TEST", "PROD"] | None = None,
        validation: bool = True,
        delay: bool = False,
        headers: dict[str, str] | None = None,
    ) -> None:
        """Make the transfer to statbanken at the end of initializing the object.

        May run the validations from the StatbankValidation class before the transfer.
        """
        self._set_user_attrs(shortuser=shortuser, cc=cc, bcc=bcc)
        self.date: datetime.date
        if isinstance(date, str):
            try:
                self.date = (
                    datetime.datetime.strptime(
                        date,
                        "%Y-%m-%d",
                    )
                    .astimezone(OSLO_TIMEZONE)
                    .date()
                )
            except ValueError as e:
                error_msg = "Skriv inn datoformen for publisering som 1900-01-01"
                raise TypeError(error_msg) from e
        elif isinstance(date, datetime.datetime):
            self.date = self.date = date.date()
        else:
            self.date = date

        self.data = data
        if not (isinstance(tableid, str) and tableid.isdigit()):
            error_msg = "Loaduser is no longer a parameter, make sure the tableid parameter is a string of digits."
            raise ValueError(error_msg)
        self.tableid = tableid
        self.overwrite = overwrite
        self.approve = _approve_type_check(approve)
        StatbankAuth.__init__(self, use_db)
        self.validation = validation
        self.__delay = delay
        self.oppdragsnummer: str = ""
        self.boundary = "12345"
        if validation:
            self._validate_original_parameters()

        self.urls = self._build_urls()
        if not self.delay:
            if headers:
                self.transfer(headers)
            else:
                self.transfer()

    def transfer(self, headers: dict[str, str] | None = None) -> None:
        """Transfers your data to Statbanken.

        Make sure you've set the publish-date correctly before sending.
        Will only work if the transfer has not already been sent, meaning it was "delayed".

        Args:
            headers (dict[str, str] | None): Mostly for internal use by the package.
                Needs to be a finished compiled headers for a request including Authorization.

        Raises:
            ValueError: If the transfer is already transferred.
        """
        # In case transfer has already happened, don't transfer again
        if self.oppdragsnummer:
            error_msg = f"Already transferred? {self.urls['gui'] + self.oppdragsnummer} Remake the StatbankTransfer-object if intentional."
            raise ValueError(error_msg)
        if headers is None:
            self.headers = self._build_headers()
        else:
            self.headers = headers
        try:
            self.params = self._build_params()
            self._validate_datatype()
            self.body = self._body_from_data()

            url_load_params = self.urls["loader"] + urllib.parse.urlencode(self.params)
            urllib.parse.urlparse(url_load_params)  # Test to see if url is valid format
            self.response = self._make_transfer_request(url_load_params)
            self._cleanup_response()
        finally:
            del self.headers  # Cleaning up auth-storing
            self.__delay = False
        self._handle_response()

    def __str__(self) -> str:
        """Print a string with the status of the transfer."""
        first_line = f"Overføring for statbanktabell {self.tableid}.\n"
        if self.delay:
            result = f"""{first_line}Ikke overført enda."""
        else:
            result = f"""{first_line}Publisering: {self.date.strftime('%d-%m-%Y,')}.\nLastelogg: {self.urls['gui'] + self.oppdragsnummer}"""
        return result

    def __repr__(self) -> str:
        """Get a representation of how to recreate the object using parameters."""
        return f'StatbankTransfer([data], tableid="{self.tableid}")'

    @property
    def delay(self) -> bool:
        """Obfuscate the delay a bit from the user. We dont want transfers transferring again without recreating the object."""
        return self.__delay

    def _set_user_attrs(
        self,
        shortuser: str = "",
        cc: str = "",
        bcc: str = "",
    ) -> None:
        if shortuser:
            self.shortuser = shortuser
        else:
            self.shortuser = os.environ["JUPYTERHUB_USER"].split("@")[0]
        if cc:
            self.cc = cc
        else:
            self.cc = self.shortuser
        if bcc:
            self.bcc = bcc
        else:
            self.bcc = self.cc

    def to_json(self, path: str = "") -> str | None:
        """Store a copy of the current state of the transfer-object as a json.

        If path is provided, tries to write to it,
        otherwise will return a json-string for you to handle like you wish.

        Args:
            path (str): if provided, will try to write a json to a local path.

        Returns:
            None: If path is provided, tries to write a json there and returns nothing.
            str: If path is not provided, returns the json-string for you to handle as you wish.
        """
        logger.warning(
            "Warning, some nested, deeper data-structures"
            " like dataframes and other class-objects will not be serialized",
        )
        json_content = json.dumps(self.__dict__, default=lambda _: "<not serializable>")
        # If path provided write to it, otherwise return the string-content
        if path:
            logger.info("Writing to %s", path)
            path_path = Path(path)
            path_path.parent.mkdir(parents=True, exist_ok=True)
            with path_path.open(mode="w") as json_file:
                json_file.write(json_content)
        else:
            return json.dumps(json_content)
        return None

    def _validate_original_parameters(self) -> None:
        for _, shortuser in enumerate([self.shortuser, self.cc, self.bcc]):
            if len(shortuser) != SSB_TBF_LEN or not isinstance(shortuser, str):
                error_msg = f'Brukeren {shortuser} - "trebokstavsforkortelse" - må være tre bokstaver...'
                raise ValueError(error_msg)

        if not isinstance(self.date, datetime.date):
            error_msg = "(datetime.date) Sett publiseringsdatoen til et gyldig datetime.date objekt."  # type: ignore[unreachable]
            raise TypeError(error_msg)

        if not isinstance(self.overwrite, bool):
            error_msg = "(Bool) Sett overwrite til enten False = ingen overskriving (dubletter gir feil), eller  True = automatisk overskriving."  # type: ignore[unreachable]
            raise TypeError(error_msg)

        if self.approve not in iter(Approve):
            error_msg = "(Integer) Sett approve til enten 0 = manuell, 1 = automatisk (umiddelbart), eller 2 = JIT-automatisk (just-in-time)"
            raise ValueError(error_msg)

    def _validate_datatype(self) -> None:
        for deltabell_name, deltabell_data in self.data.items():
            if not isinstance(deltabell_name, str):
                error_msg = f"{deltabell_name} is not a string."  # type: ignore[unreachable]
                raise TypeError(error_msg)
            if not isinstance(deltabell_data, pd.DataFrame):
                error_msg = f"Data for {deltabell_name}, must be a pandas DataFrame"  # type: ignore[unreachable]
                raise TypeError(error_msg)

    def _body_from_data(self) -> str:
        # Data should be a iterable of pd.DataFrames at this point,
        # reshape to body
        body = ""
        for filename, elem in self.data.items():
            # Replace all nans in data
            elem_fillna = elem.copy().astype("string").fillna("")
            body += f"--{self.boundary}"
            body += f"\nContent-Disposition:form-data; filename={filename}"
            body += "\nContent-type:text/plain\n\n"
            csv_content = elem_fillna.to_csv(sep=";", index=False, header=False)
            body += str(csv_content)
        body += f"\n--{self.boundary}--"
        return body.replace("\n", "\r\n")  # Statbank likes this?

    def _build_params(self) -> dict[str, str]:
        return {
            "initialier": self.shortuser,
            "hovedtabell": self.tableid,
            "publiseringsdato": self.date.strftime("%Y-%m-%d"),
            "fagansvarlig1": self.cc,
            "fagansvarlig2": self.bcc,
            "auto_overskriv_data": str(int(self.overwrite)),
            "auto_godkjenn_data": str(int(self.approve)),
        }

    def _make_transfer_request(
        self,
        url_params: str,
    ) -> r.Response:
        result = r.post(url_params, headers=self.headers, data=self.body, timeout=100)
        # Trying to clean all auth etc out of response
        try:
            result.raise_for_status()
        except r.HTTPError:
            logger.error(result.text)
            raise
        return result

    def _cleanup_response(self) -> None:
        if hasattr(self.response.request, "headers"):
            del (
                self.response.request.headers
            )  # Auth is stored here also, for some reason
        if hasattr(self.response, "cookies"):
            del self.response.cookies
        if hasattr(self.response, "raw"):
            del self.response.raw
        gc.collect()  # Hoping this removes the del-ed stuff from memory

    def _handle_response(self) -> None:
        pattern_work_number = re.compile(r"lasteoppdragsnummer:(\d+)")
        pattern_publish_date = re.compile(
            r"Publiseringsdato '(\d{2}.\d{2}.\d{4} \d{2}:\d{2}:\d{2})'",
        )
        pattern_publish_time = re.compile(r"Publiseringstid '(\d{2}):(\d{2})'")

        resp_json: TransferResultType = self.response.json()
        response_msg = resp_json["TotalResult"]["Message"]
        match_oppdragsnummer = cast(
            "re.Match[str]",
            pattern_work_number.search(response_msg),
        )
        match_publish_date = cast(
            "re.Match[str]",
            pattern_publish_date.search(response_msg),
        )
        match_publish_time = cast(
            "re.Match[str]",
            pattern_publish_time.search(response_msg),
        )

        if not match_oppdragsnummer[1].isdigit():
            error_msg = f"Lasteoppdragsnummer: {match_oppdragsnummer[1]} er ikke ett rent nummer."
            raise ValueError(error_msg)

        self.oppdragsnummer = match_oppdragsnummer[1]
        publish_date = datetime.datetime.strptime(
            match_publish_date[1],
            r"%d.%m.%Y %H:%M:%S",
        ).astimezone(OSLO_TIMEZONE)
        publish_hour, publish_minute = map(int, match_publish_time.groups())
        publish_date = publish_date.replace(hour=publish_hour, minute=publish_minute)
        logger.info("Publisering satt til: %s", publish_date.isoformat("T", "seconds"))
        logger.info(
            "Følg med på lasteloggen (tar noen minutter): %s",
            {self.urls["gui"] + self.oppdragsnummer},
        )
        self.resp_json = resp_json



from __future__ import annotations

import datetime as dt
import enum
import zoneinfo


class Approve(enum.IntEnum):
    """Enum for approval codes."""

    MANUAL = 0
    """Manual approval."""
    AUTOMATIC = 1
    """Automatic approval at transfer-time (immediately)."""
    JIT = 2
    """Just in time approval right before publishing time."""


def _approve_type_check(approve: Approve | int | str) -> Approve:
    result: Approve

    match approve:
        case Approve():
            result = approve
        case int():
            result = Approve(approve)
        case str() if approve.isdigit():
            result = Approve(int(approve))
        case str():
            result = getattr(Approve, approve.upper())
        case _:
            error_msg = f"Dont know how to handle approve of type {type(approve)}"  # type: ignore[unreachable]
            raise TypeError(error_msg)

    return result


OSLO_TIMEZONE = zoneinfo.ZoneInfo("Europe/Oslo")
TOMORROW = dt.date.today() + dt.timedelta(days=1)  # noqa: DTZ011
APPROVE_DEFAULT_JIT = Approve.JIT
STATBANK_TABLE_ID_LEN = 5
REQUEST_OK = 200
SSB_TBF_LEN = 3



from __future__ import annotations

import re
import urllib
from collections import Counter
from http import HTTPStatus
from typing import TYPE_CHECKING
from typing import Any
from typing import TypeVar

import requests as r

if TYPE_CHECKING:
    from collections.abc import Iterable
    from collections.abc import Sequence

    import pandas as pd

    from .api_types import QueryPartType
    from .api_types import QueryWholeType

from .api_exceptions import StatbankParameterError
from .api_exceptions import StatbankVariableSelectionError
from .api_exceptions import TooBigRequestError
from .response_to_pandas import response_to_pandas
from .statbank_logger import logger

# Getting data from Statbank

STATBANK_TABLE_ID_LENGTH = 5
T = TypeVar("T")


def apidata(
    id_or_url: str = "",
    payload: QueryWholeType | None = None,
    include_id: bool = False,
) -> pd.DataFrame:
    """Get the contents of a published statbank-table as a pandas Dataframe, specifying a query to limit the return.

    Args:
        id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".
        payload (QueryWholeType | None): a dict in the shape of a QueryWhole, to include with the request, can be copied from the statbank-webpage.
        include_id (bool): If you want to include "codes" in the dataframe, set this to True

    Returns:
        pd.DataFrame: The table-content

    Raises:
        ValueError: If the first parameter is not recognized as a statbank ID or a direct url.
    """
    if payload is None:
        payload_now: QueryWholeType = {
            "query": [],
            "response": {"format": "json-stat2"},
        }
    else:
        payload_now = payload
    if len(id_or_url) == STATBANK_TABLE_ID_LENGTH and id_or_url.isdigit():
        url = f"https://data.ssb.no/api/v0/no/table/{id_or_url}/"
    else:
        test_url = urllib.parse.urlparse(id_or_url)
        if not all([test_url.scheme, test_url.netloc]):
            error_msg = (
                "First parameter not recognized as a statbank ID or a direct url"
            )
            raise ValueError(error_msg)
        url = id_or_url

    logger.info(url)
    # Spør APIet om å få resultatet med requests-biblioteket
    resultat = r.post(url, json=payload_now, timeout=10)
    if not resultat.ok:
        _read_error(id_or_url, payload_now, resultat)

    # Få pd.DataFrame fra resultatet
    table_data = response_to_pandas(resultat, include_id=include_id)

    return table_data.convert_dtypes()


def apidata_all(id_or_url: str = "", include_id: bool = False) -> pd.DataFrame:
    """Get ALL the contents of a published statbank-table as a pandas Dataframe.

    Args:
        id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".
        include_id (bool): If you want to include "codes" in the dataframe, set this to True

    Returns:
        pd.DataFrame: Table-content
    """
    return apidata(id_or_url, apidata_query_all(id_or_url), include_id=include_id)


def apimetadata(id_or_url: str = "") -> dict[str, Any]:
    """Get the metadata of a published statbank-table as a dict.

    Args:
        id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".

    Returns:
        dict[str, Any]: The metadata of the table as the json returned from the API-get-request.

    Raises:
        ValueError: If the first parameter is not recognized as a statbank ID or a direct url.
    """
    if len(id_or_url) == STATBANK_TABLE_ID_LENGTH and id_or_url.isdigit():
        url = f"https://data.ssb.no/api/v0/no/table/{id_or_url}/"
    else:
        test_url = urllib.parse.urlparse(id_or_url)
        if not all([test_url.scheme, test_url.netloc]):
            error_msg = (
                "First parameter not recognized as a statbank ID or a direct url"
            )
            raise ValueError(error_msg)
        url = id_or_url
    res = r.get(url, timeout=5)
    res.raise_for_status()
    meta: dict[str, Any] = res.json()
    return meta


def apicodelist(
    id_or_url: str = "",
    codelist_name: str = "",
) -> dict[str, str] | dict[str, dict[str, str]]:
    """Get one specific or all the codelists of a published statbank-table as a dict or nested dicts.

    Args:
        id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".
        codelist_name (str): The name of the specific codelist to get.

    Returns:
        dict[str, str] | dict[str, dict[str, str]]: The codelist of the table as a dict or a nested dict.

    Raises:
        ValueError: If the specified codelist_name is not in the returned metadata.
    """
    metadata = apimetadata(id_or_url)
    results = {}
    for col in metadata["variables"]:
        results[col["code"]] = dict(zip(col["values"], col["valueTexts"], strict=False))
    if codelist_name == "":
        return results
    if codelist_name in results:
        return results[codelist_name]
    for col in metadata["variables"]:
        if codelist_name == col["text"]:
            return dict(zip(col["values"], col["valueTexts"], strict=False))
    col_names = ", ".join([col["code"] for col in metadata["variables"]])
    error_msg = f"Cant find {codelist_name} among the available names: {col_names}"
    raise ValueError(error_msg)


def apidata_query_all(id_or_url: str = "") -> QueryWholeType:
    """Builds a query for ALL THE DATA in a table based on a request for metadata on the table.

    Args:
        id_or_url (str): The id of the STATBANK-table to get the total query for, or supply the total url, if the table is "internal".

    Returns:
        QueryWholeType: The prepared query based on all the codes in the table.
    """
    meta = apimetadata(id_or_url)["variables"]
    code_list: list[QueryPartType] = []
    for code in meta:
        tmp: QueryPartType = {"code": "", "selection": {"filter": "", "values": []}}
        for k, v in code.items():
            if k == "code":
                tmp["code"] = v
            if k == "values":
                tmp["selection"] = {"filter": "item", "values": v}
        code_list += [tmp]
    return {"query": code_list, "response": {"format": "json-stat2"}}


# Credit: https://github.com/sehyoun/SSB_API_helper/blob/master/src/ssb_api_helper.py
def apidata_rotate(
    df: pd.DataFrame,
    ind: str = "year",
    val: str = "value",
) -> pd.DataFrame:
    """Rotate the dataframe so that years are used as the index.

    Args:
        df (pd.DataFrame): dataframe (from <get_from_ssb> function)
        ind (str): string of column name denoting time
        val (str): string of column name denoting values

    Returns:
        pd.DataFrame: pivoted dataframe
    """
    return df.pivot_table(
        index=ind,
        values=val,
        columns=[i for i in df.columns if i not in (ind, val)],
    )


# Error handeling and HTTP-return code interpretations follow
def _list_up(sequence: Sequence[str], conjunction: str = "and") -> str:
    if len(sequence) == 1:
        return sequence[0]

    return f"{', '.join(sequence[:-1])} {conjunction} {sequence[-1]}"


def _find_duplicates(items: Iterable[T]) -> list[T]:
    return [item for item, n in Counter(items).items() if n > 1]


def _read_error(id_or_url: str, query: QueryWholeType, response: r.Response) -> None:
    """Raises an appropriate error."""
    error_message: str | None

    if response.status_code == HTTPStatus.FORBIDDEN:
        error_message = "Your query is too big. The API is limited to 800,000 cells (incl. empty cells)"
        raise TooBigRequestError(error_message)

    if response.status_code == HTTPStatus.BAD_REQUEST:
        api_error_message = response.json().get("error", "")

        match = re.match(
            r"The request for variable '(?P<variable>.+)' has an error\. Please check your query\.",
            api_error_message,
        )

        if match:
            variable = match["variable"]
            error_message = _check_selection(variable, id_or_url, query)
            if not error_message:
                error_message = (
                    f'Your query failed with the error message: "{api_error_message}"'
                )
            raise StatbankVariableSelectionError(error_message)

        error_message = (
            f'Your query failed with the error message: "{api_error_message}"'
        )
        raise StatbankParameterError(error_message)

    response.raise_for_status()


def _check_selection(
    variable: str,
    id_or_url: str,
    query: QueryWholeType,
) -> str | None:
    """Checks for common errors in your query selection, and returns a error message.

    When the query fails with the message "The request for variable ... has an error,"
    check that the selection don't contains duplicate and invalid values against the metadata and
    check that the filter is set to "all" when selecting with a wildcard (*).
    Metadata for aggregations are not available, so we can't inspect selection with agg and agg_single filters.
    Errors with "top" and "all" filter always fail with "parameter error", so we don't validate them here.
    """
    query_part = next(filter(lambda part: part["code"] == variable, query["query"]))
    query_selection = query_part["selection"]

    match = re.fullmatch(
        r"(?P<filtertype>(item)|(vs|agg|agg_single))(?(3):(?P<aggregering>.+))",
        query_selection["filter"],
    )

    if match is None:
        return f'The filter don\'t match one of the types "item", "all", "top", "vs:", "agg:", "agg_single:", for variable {variable}.'

    filter_type = match["filtertype"]

    duplicates = _find_duplicates(query_selection["values"])
    if len(duplicates) > 0:
        return (
            f"The value(s) {_list_up(duplicates)} is duplicated for variable {variable}"
        )

    if any("*" in values for values in query_selection["values"]):
        return (
            f"One of the values for the variable {variable} contains a wildcard character (*)."
            'If you wish to select several or all values with a wildcard, "filter" must be set to "all"'
        )

    if filter_type in ("agg", "agg_single"):
        return (
            "A value is probably invalid for variable {variable}, but an aggregation is used,"
            "and metadata for aggregations is not available, so it is not possible to determine witch."
        )

    meta = apimetadata(id_or_url)

    variable_meta = next(
        filter(lambda part: part["code"] == variable, meta["variables"]),
    )

    invalid_values = [
        value
        for value in query_selection["values"]
        if value not in variable_meta["values"]
    ]

    if len(invalid_values) > 0:
        return f"Invalid value(s) {_list_up(invalid_values)} have been specified for the variable {variable}"

    return None



import json
from typing import TYPE_CHECKING
from typing import Any

import numpy as np
import pandas as pd
import requests

if TYPE_CHECKING:
    from numpy.typing import NDArray


class StatbankDataBuildError(Exception):
    """Custom exception for errors that occur during the building of Statbank data."""


def stack_categories(varcodes: dict[str, list[str]]) -> pd.DataFrame:
    """Stacks the categories of the variables (in the correct order) from the response into a DataFrame.

    Args:
        varcodes (dict[str, list[Any]]): A dictionary where the keys are variable names and the values are lists of codes.

    Returns:
        pd.DataFrame: A pandas DataFrame containing the stacked categories.
    """
    lenghts: list[int] = [len(codes) for codes in varcodes.values()]

    nrepeats: int = int(np.prod(lenghts))
    ntiles: int = 1

    data_dict: dict[str, NDArray[Any]] = {}

    var: str
    codes: list[str]

    for var, codes in varcodes.items():
        ncat: int = len(codes)

        # reduce nrepeats before creating the array
        nrepeats = int(nrepeats / ncat)

        # create array
        data_dict[var] = np.repeat(np.tile(codes, reps=ntiles), repeats=nrepeats)

        # reduce ntiles afterwards
        ntiles = int(ncat * ntiles)

    return pd.DataFrame(data_dict)


def response_to_pandas(
    response: requests.models.Response,
    include_id: bool = False,
) -> pd.DataFrame:
    """Converts a response from the SSB Statbank API to a pandas DataFrame.

    Args:
        response (requests.models.Response): The response object from the API.
        include_id (bool): Whether to include the ID columns in the output. Defaults to False.

    Returns:
        pd.DataFrame: A pandas DataFrame containing the data from the response.

    Raises:
        StatbankDataBuildError: If the number of categories and values do not match.
    """
    content: dict[str, Any] = json.loads(response.content.decode("UTF-8"))

    dimension: dict[str, Any] = content["dimension"]

    varnames: dict[str, str] = {}
    codes: dict[str, list[str]] = {}
    labels: dict[str, dict[str, str]] = {}

    var: str
    var_info: dict[str, Any]

    for var, var_info in dimension.items():
        varnames[var] = var_info["label"]
        codes[var] = list(var_info["category"]["index"].keys())
        labels[var] = var_info["category"]["label"]

    data: pd.DataFrame = stack_categories(codes)
    value: list[str | float | int] = content["value"]

    if data.shape[0] != len(value):
        message: str = (
            "The number of categories and values do not match! This is likely a bug!"
        )
        raise StatbankDataBuildError(message)

    data["value"] = value

    id_col: str
    id_col2: str
    codes2labels: dict[str, str]

    for id_col, codes2labels in labels.items():
        label_col: str = varnames[id_col]

        # probably a prettier way of doin this
        # linter also complains if we overwrite id_col (id_col = "id_" + id_col)
        if id_col == label_col:  # not sure if this is possible, but just in case
            data = data.rename(columns={id_col: "id_" + id_col})
            id_col2 = "id_" + id_col
        else:
            id_col2 = id_col

        data[label_col] = data[id_col2].map(codes2labels)

    if not include_id:  # early return
        cols: list[str] = [*list(varnames.values()), "value"]
        return data[cols]

    knitted_cols: list[str] = []

    for id_col, label_col in varnames.items():
        # ugly doing this twice
        # linter also complains if we overwrite id_col (id_col = "id_" + id_col)
        id_col2 = "id_" + id_col if id_col == label_col else id_col
        knitted_cols += [id_col2, label_col]

    knitted_cols += ["value"]

    return data[knitted_cols]



class StatbankApiError(Exception):
    """Base class for Statbank Api Errors."""


class TooBigRequestError(StatbankApiError):
    """Exception for request that is to big to handle for the Statbank Api.

    The API is limited to 800,000 cells (incl. empty cells)
    """


class StatbankParameterError(StatbankApiError):
    """Exception for invalid request."""


class StatbankVariableSelectionError(StatbankApiError):
    """Exception for invalid variable selection."""
